// 2010 Mike Brent aka Tursi aka HarmlessLion.com
// This software is provided AS-IS. No warranty
// express or implied is provided.
// http://harmlesslion.com - visit the web page for contact info
//
// Wrapper for reSID emulation device
// SID card was created by Marc Hull. 
// The SID chip is clocked at 1MHz using a distinct clock from the TI's
// Data is passed into the SID using latches rather than buffers, so a
// value once written continues to be presented to the chip until
// it is changed. The way to clear it is to write to one of the
// SID's read-only registers (this prevents it from affecting cycles).
// Rather than a traditional CRU system, the SID card switches itself
// in when a CRU operation under >0FFF occurs (ie: keyboard!). This
// allows the card to be active during Extended BASIC, for instance.
// The registers are mapped from >5800 to >5836 on even bytes only.
// The read/write line is fixed so it is not possible to read from
// the SID chip.
// Also, when any CRU address above >0FFF is accessed, the SID card
// is switched out and the SID RESET is asserted!
//SID REGISTER ADDRESSING

//SID REG.	REAL HEX	REAL DEC	C64 LOC		MEANING
//________	___________	__________	________	__________________
//0			5800		22528		54272		V1 FREQ LO
//1			5802		22530		54273		V1 FREQ HI
//2			5804		22532		54274		V1 PULSE LO
//3			5806		22534		54275		V1 PULSE HI
//4			5808		22536		54276		V1 WAVEFORM
//5			580A		22538		54277		V1 ATTAK/DECAY
//6			580C		22540		54278		V1 SUSTAIN/RELEASE


//7			580E		22542		54279		V2 FREQ LO
//8			5810		22544		54280		V2 FREQ HI
//9			5812		22546		54281		V2 PULSE LO
//10		5814		22548		54282		V2 PULSE HI
//11		5816		22550		54283		V2 WAVEFORM
//12		5818		22552		54284		V2 ATTAK/DECAY
//13		581A		22554		54285		V2 SUSTAIN/RELEASE


//14		581C		22556		54286		V3 FREQ LO
//15		581E		22558		54287		V3 FREQ HI
//16		5820		22560		54288		V3 PULSE LO
//17		5822		22562		54289		V3 PULSE HI
//18		5824		22564		54290		V3 WAVEFORM
//19		5826		22566		54291		V3 ATTAK/DECAY
//20		5828		22568		54292		V3 SUSTAIN/RELEASE

//21		582A		22570		54293		ALL FILTER LO
//22		582C		22572		54294		ALL FILTER HI
//23		582E		22574		54295		ALL FILTER RESONANCE
//24		5830		22576		54296		ALL VOLUME

//25		5832		22578		54297		READ ONLY N/A  POTX
//26		5834		22580		54298		READ ONLY N/A  POTY
//27		5834		22582		54299		READ ONLY N/A  OSC VALUE
//28		5836		22584		54300		READ ONLY N/A  ENVELOPE OF VOICE 3

#include <windows.h>
#include "sid.h"

typedef unsigned __int8 UINT8;
typedef unsigned __int8 Byte;
typedef unsigned __int16 Word;
typedef unsigned __int32 DWord;

#define SID_CLOCK 1000000;

int nOutFrequency=22050;				// output frequency - overridden by user

// define the SID object
SID mySid;

// latched data
Byte nCachedReg = 0xff;
Byte nCachedData = 0xff;

bool sid_blaster_enable = true;
bool bSidBankedIn = false;

void InitSid() {
	OutputDebugString("SIDDll, based on reSID 0.9, GPL code.\n");
	mySid.reset();
}

void SetSidFrequency(int nNewFreq) {
	nOutFrequency = nNewFreq;
	InitSid();
}

void SetSidEnable(bool en) {
	sid_blaster_enable = en;
	if (en) {
		OutputDebugString("SID is enabled\n");
	} else {
		OutputDebugString("SID is disabled\n");
	}
}

bool GetSidEnable() {
	return sid_blaster_enable;
}

void SetSidBanked(bool b) {
	bSidBankedIn = b;
}

// fill the output audio buffer with signed 16-bit values from the SID chip
// nAudioIn contains a fixed value to add to all samples (though not wired up on this chip)
// (this emu doesn't run speech through there, though, speech gets its own buffer for now)
void sid_update(short *buf, double nAudioIn, int nSamples) {
	if ((!bSidBankedIn) || (!sid_blaster_enable)) {
		// if the SID is not banked in, then we are in continuous reset
		mySid.reset();
		memset(buf, 0, nSamples*2);		// sizeof short
	} else {
		int nClock = SID_CLOCK;		// override the 9919 clock variable

		// nClock is the input clock frequency, exactly in this case
		// nOutFrequency is the frequency at which we actually output bytes
		double nTimePerClock=1.0/(nClock);
		double nTimePerSample=1.0/(double)nOutFrequency;
		int nClocksPerSample = (int)(nTimePerSample / nTimePerClock + 0.5);		// +0.5 to round up if needed

		while (nSamples) {
			// run chip - we have to run every clock because
			// we need to refresh the latch each time
			int cnt = nClocksPerSample;
			// sanity check - something things seem to run away?
			if (cnt > 1000) {
				cnt=1000;	// that's a full second!
			}
			while (cnt--) {
				mySid.write(nCachedReg, nCachedData);
				mySid.clock();
			}

			// write sample
			nSamples--;
			short nSample = (short)mySid.output();
			*(buf++)=nSample; 
		}
	}
}

// takes the raw TI address in
// there's no read function because the SID card doesn't implement read
void write_sid(Word ad, Byte dat) {
	// cache the values - the way the card is hooked up, the write will happen every cycle
	if ((ad > 0x5800) && (ad<0x6000) && ((ad & 0x0001) == 0)) {
		ad &= 0x003f;
		nCachedReg = (Byte)(ad>>1);
		nCachedData = dat;

		// and write it now
		mySid.write(nCachedReg, nCachedData);
	}

	// else out of range or not a valid number
}
